/*
 * cooking/cooking.go --
 *
 *      Solution of the student cooking problem in go using a
 *      self-made Barrier object.
 */

package main

import (
	"fmt"
	goopt "github.com/droundy/goopt"
)

var n = goopt.Int([]string{"-n", "--students"}, 5, "number of students")
var c = goopt.Int([]string{"-c", "--cooks"}, 1, "number of cooks")

// A Barrier represents a barrier that opens once a certain number
// of objects have reached the barrier.
type Barrier struct {
	n   int       // number of objects needed to pass
	c   int       // current number of objects waiting
	in  chan bool // channel to signal arrival
	out chan bool // channel to signal departure
}

func NewBarrier(n int, descr string) *Barrier {
	b := &Barrier{n: n, c: 0, in: make(chan bool), out: make(chan bool)}
	go func() {
		for {
			for b.c < b.n {
				<-b.in
				b.c++
			}
			for b.c > 0 {
				b.out <- true
				b.c--
			}
		}
	}()
	return b
}

func (b *Barrier) Pass(cont bool) {
	b.in <- cont
	<-b.out
}

var barrier_all_students *Barrier
var barrier_all_cooks *Barrier
var barrier_all_lazy *Barrier

const (
	STUDENT_TYPE_COOK = 'C'
	STUDENT_TYPE_LAZY = 'L'

	STUDENT_STATE_IDLE     = ' '
	STUDENT_STATE_ARRIVING = 'A'
	STUDENT_STATE_COOKING  = 'C'
	STUDENT_STATE_SITTING  = 'S'
	STUDENT_STATE_EATING   = 'E'
	STUDENT_STATE_CLEANING = 'W'
)

type Student struct {
	s_id    int
	s_state int
	s_type  int
}

func (student *Student) State(new_state int) {
	fmt.Printf("%3d: %c -> %c\n", student.s_id, student.s_state, new_state)
	student.s_state = new_state
}

func (student *Student) Party(done chan bool) {
	student.State(STUDENT_STATE_ARRIVING)
	barrier_all_students.Pass(true)
	if student.s_type == STUDENT_TYPE_COOK {
		student.State(STUDENT_STATE_COOKING)
		barrier_all_cooks.Pass(false)
	}
	student.State(STUDENT_STATE_SITTING)
	barrier_all_students.Pass(true)
	student.State(STUDENT_STATE_EATING)
	barrier_all_students.Pass(true)
	if student.s_type != STUDENT_TYPE_COOK {
		student.State(STUDENT_STATE_CLEANING)
		barrier_all_lazy.Pass(false)
	}
	student.State(STUDENT_STATE_IDLE)
	barrier_all_students.Pass(false)
	done <- true
}

func run(n int, c int) {

	barrier_all_students = NewBarrier(n, "all")
	barrier_all_cooks = NewBarrier(c, "cooks")
	barrier_all_lazy = NewBarrier(n-c, "lazy")
	done := make(chan bool)

	students := make([]Student, n)
	for i := 0; i < n; i++ {
		students[i].s_id = i
		students[i].s_state = STUDENT_STATE_IDLE
		if i < c {
			students[i].s_type = STUDENT_TYPE_COOK
		} else {
			students[i].s_type = STUDENT_TYPE_LAZY
		}
		go students[i].Party(done)
	}

	for i := 0; i < n; i++ {
		<-done
	}
}

func main() {
	goopt.Parse(nil)
	if *c > *n {
		panic(fmt.Sprintf("%d students can't have %d cooks", *n, *c))
	}
	run(*n, *c)
}
